module Day2.Shared (
  Shape (Rock, Paper, Scissors),
  shapeScore,
  Score (Score),
  Game (Game),
  gameWin,
  gameLoss,
  gameDraw,
  play,
  winAgainst,
  loseAgainst
) where

data Shape = Rock | Paper | Scissors
  deriving (Show, Eq, Enum)

newtype Score = Score Int
  deriving (Show, Eq)

instance Semigroup Score where
  (<>) (Score x) (Score y) = Score (x + y)

instance Monoid Score where
  mempty = Score 0

gameWin :: Score
gameWin = Score 6

gameDraw :: Score
gameDraw = Score 3

gameLoss :: Score
gameLoss = Score 0

newtype Game = Game (Shape, Shape)
  deriving (Show, Eq)

shapeScore :: Shape -> Score
shapeScore = Score . (+1) . fromEnum

winAgainst :: Shape -> Shape
winAgainst Rock = Paper
winAgainst Paper = Scissors
winAgainst Scissors = Rock

loseAgainst :: Shape -> Shape
loseAgainst Rock = Scissors
loseAgainst Paper = Rock
loseAgainst Scissors = Paper


play :: Game -> Score
play (Game (x, y))
 | winAgainst  x == y = gameWin  <> shapeScore y
 | loseAgainst x == y = gameLoss <> shapeScore y
 | otherwise          = gameDraw <> shapeScore y
